/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <float.h>

#include <packet.h>
#include <eventhandler.h>
#include <scheduler.h>
#include <debug.h>

//Event class: Any event that needs to be scheduled must be an event class.
//Eventhandlers register events to be scheduled with the scheduler.
//The scheduler formats the event as an event object and inserts it 
//into its scheduling list.
//The event object contains a pointer to the event handler that registered it.
//It may also contain a pointer to a packet, if the event is related to 
//packet processing.


//Function to compare two timestamps (doubles).
//Comparing doubles should be done by choosing a level of significance.
//The level of significance here is 1 picosecond.
//This is enought to support chipping rates of 1 Tb/s.
//TBD: The comparison may also be done by converting to integer upto 
//the level of significance and then comparing.

inline int areequal(double d1, double d2)
{
//The epsilon is 1 picosecond
	double epsilon = 	1E-12;
	if(fabs(d1 - d2) <= epsilon)
		return 1;
	else
		return 0;
}

int Scheduler:: insert(Event *e1, Event *e2)
{
//Insert e2 after e1
	if (e1 == NULL)  //Insert as head node
	{
		dprintf("scheduler.cc: Schedule::insert(): Event due before head\n");	

		if(head != NULL)  //There may already be a head node
		{
			e2->next = head;	
			head->prev = e2;
		}
		else							//There was no head node
			e2->next  = NULL;

		head = e2;
		e2->prev = NULL;
		return 1;
	}



	e2->next = e1->next;

	if(e1->next != NULL)
	{
		dprintf("scheduler.cc: Schedule::insert(): Event due between other events\n");	
		e1->next->prev = e2;
	}
	else
	{
		dprintf("scheduler.cc: Schedule::insert(): Event due after tail\n");	
	}

	e1->next= e2;
	e2->prev = e1;
	return 1;

}

double Scheduler::start = 0;

Scheduler::Scheduler() : head(NULL)
{
	debug = debugon("scheduler");
	currenttime = 0;
	halt = 0;
}



Scheduler::~Scheduler()
{


}

//Main scheduler function. Removes the first eventhandler from the 
//eventhandler queue and calls its handleevent() function. Updates 
//the current time.
int Scheduler:: run()
{
	dprintf("scheduler.cc: Schedule::run(): Entered function\n");



	if(!head) 
		return 0;

//Initialize the starting time.
	static double start = head->when;

	//Set current time to the event time.
	currenttime = head->when;	
	

	Event *tmphead = head;	
	//Remove event from queue, if no more events to schedule return FALSE
	if(head->next == NULL)
	{
		head = NULL;
	}
	else
	{
		head = head->next;
		head->prev = NULL;
	}

	dprintf("tmp : %d\n", tmphead);
	dprintf("eh : %d\n", tmphead->eh);
	//Call event handler, user is responsible for deleting head.
	tmphead->eh->handleevent(tmphead->p);
	delete tmphead; 
	return 1;
}


//Event handling registration function. Event handler's register themselves
//with the scheduler in order to be scheduled. The scheduleevent function 
//adds the eventhandler to the queue in the appropriate place. The queue 
//is maintained in the order of scheduling.
void Scheduler:: scheduleevent(EventHandler *eh, Packet *p, double delay )
{




	dprintf("scheduler.cc: Schedule::scheduleevent(): Entered function\n");	
	if(halt != 0)
	{
		dprintf("scheduler.cc: Schedule::scheduleevent(): Scheduler halted. No more new events will be scheduled. Pending events will be scheduled\n");
		return;
	}
	
	Event *ev = new Event();
	if(ev == NULL)
	{
		printf("Error: scheduler.cc: Memory allocation failed\n");
		exit(-1);
	}
	ev->eh = eh;
	ev->p = p;
//Store times as absolute times.
	ev->when  = currenttime + delay;



	dprintf("scheduler.cc: Schedule::scheduleevent(): Current time: %.12f\n", currenttime);
	dprintf("scheduler.cc: Schedule::scheduleevent(): Delay: %.12f\n", delay);
	dprintf("scheduler.cc: Schedule::scheduleevent(): Scheduling delay: %.12f\n", ev->when);



	//There are no elements in the queue.
	if(head == NULL)
	{
		dprintf("scheduler.cc: Schedule::scheduleevent(): No elements in the queue\n");	
		dprintf("scheduler.cc: Schedule::scheduleevent(): Scheduling event for: %.12f\n", ev->when);
		insert(head, ev);		
		return;
	}


	Event *e = head;
	dprintf("scheduler.cc: Schedule::scheduleevent(): Scheduling event for (eh): %.12f\n", ev->when);
	

	while(e != NULL)
	{
		dprintf("scheduler.cc: Schedule::scheduleevent(): Event scheduled for (e): %.12f\n", e->when);
		dprintf("scheduler.cc: Schedule::scheduleevent(): Event priority (e): %d\n", e->priority);


//Note: If there is already an event scheduled at the same time as the 
//event to be scheduled, then the event to be scheduled is added 
//infront/behind it depending on the relative priorities.

//Check if two events need to be scheduled at the same time
		if(areequal(ev->when, e->when))
		{
			if(ev->priority <= e->priority)
			{
				dprintf("scheduler.cc: Schedule::scheduleevent(): Event due at the same time, different priority\n");
				if(e->next == NULL)
				{
					dprintf("scheduler.cc: Schedule::scheduleevent(): Event due after tail event\n");	
					insert(e, ev);	
					break;
				}
				e=e->next;
				continue;
			}
			else
			{
				dprintf("scheduler.cc: Schedule::scheduleevent(): Event due at the same time, same priority\n");

				Event *first = e;
				Event *etemp = e;
				int ecount = 0;

//Count the number of event of the same priority, 
//scheduled at the same time.
				while ((etemp->priority == first->priority ) &&(etemp->next != NULL))
				{
					etemp = etemp->next;
					ecount++;
				}

				dprintf("scheduler.cc: Schedule::scheduleevent(): Found %d events of the same priority scheduled at the same time\n", ecount);	

#ifdef SCHEDULING_RANDOM
//Select a random number from 0 to ehcount and insert the new event in that position.
				int rnd = (int)(floor((random()/(pow(2, 31) -1)) *ecount));
				dprintf("scheduler.cc: Schedule::scheduleevent(): Inserting event after %d events\n", rnd);

				for(int count=0; count < rnd; count++)
					e=e->next;
				insert (e->prev, ev);
				break;

#else
				insert (e->prev, ev);
				break;
#endif

			}
		}
//End of code for events scheduled at the same time


		if(ev->when < e->when) 
		{
			dprintf("scheduler.cc: Schedule::scheduleevent(): Event due between other events\n");	
			insert(e->prev, ev);
			break;
		}

		if(e->next == NULL)
		{
			dprintf("scheduler.cc: Schedule::scheduleevent(): Event due after tail event\n");	
			insert(e, ev);	
			break;
		}
		e = e->next;
	}
}

void Scheduler::dprintf(char * format, ...)
{
	if(debug == 1)
	{
		printf("%.12f:", currenttime);
		va_list ap;
		va_start(ap, format);
		vprintf(format, ap);
		fflush(stdout);
	}
}
