/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <state.h>
#include <statshandler.h>

extern Scheduler * sched;
extern InTransit * it;
extern char outputfile[20];
extern Codeset *codeset;
extern Receiver *rcvr;
extern TrGen *trgen;
extern int propSpeed;
extern StatsHandler * stats;


//The trafic dump function.
//This class collects packets from the transmitter, 
//waits till they have passed the coupler. 
//and then passes it on to the receiver.
//dumps them to a file and collects statistics

//Note: A packet can be checked for packet errors anytime 
//after the packet's tail passes the coupler.
//This is because after the coupler, no other new packets can overlap
//with it after it passes the coupler.

//However, even after a packet's tail passes  the coupler, it can still affect
//other packets which it overlaps with but which have not passed the coupler.
 
//Note: the term headend is used synonymously with coupler in the code

extern int steadystate;

TrDump::TrDump():EventHandler()
{
	debug = debugon("trdump");
	fout = fopen(outputfile, "w");
	if (fout == NULL)
	{
		dtprintf("trgen.cc: TrDump::TrDump(): Could not open output file : %s for writing\n", outputfile);
		exit(-1);
	}
	totalpktssent = 0;
	totalbytessent = 0;
	errorpkts = 0;
	errorbytes = 0;
}

TrDump::~TrDump()
{
	fclose(fout);
}


void TrDump::process(Packet *p)
{
	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TrDump::process(): Scheduling a packet: %d\n", p->id); 		
	it->dump();	
	 	
	//Distance to headend measured in meters
	double transitdistance = p->depx - 0;
	double transittime = transitdistance/propSpeed;		
	double packetlength = (double)p->symsize * codeset->N;
	double transmissiontime = packetlength/p->chiprate;
	double totaltime = transittime + transmissiontime;
	double tailtime = p->departuretime+ totaltime;
	double delay = tailtime - sched->currenttime;
		
	dtprintf("trdump.cc: TrDump::process(): Packet id: %d Delay: %.12f s\n", p->id, delay); 		
	dprintf("\n");
	sched->scheduleevent(this, p, delay);	

}




//The event handling function of the TrDump. It dequeues a packet and
//sends it to the next component.
void TrDump::handleevent(Packet *rp)
{
	dtprintf("trdump.cc: TrDump::handleevent(): Entered function: %d\n", rp->id);
	dtprintf("trdump.cc: TrDump::handleevent(): Current time: %.12f\n", sched->currenttime);
			

	rp->status = PACKET_OK;
	PacketState *pstate = new PacketState(rp);
	if(pstate == NULL)
	{
		printf("Error: trdump.cc: Memory allocation failed\n");
		exit(-1);
	}
	pstate->ConstructStateSections(rp);
	dtprintf("trdump.cc: TrDump::handleevent(): Checking each section\n"); 		
	Section * s = pstate->head;

//Stats collection: Calculate the difference between the state estimate 
//(stored in the packet) and the actual state which the head of the packet experienced 
//at the coupler. 
	stats->calcstatediff(rp);

	int sectioncount =0;
	int totalstatechanges = 0;
	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double bitlength = chiplength * codeset->N;
	while(s != NULL)
	{
		double sectionsize = s->tail - s->head;
		dtprintf("trdump.cc: TrDump::handleevent():Section size: %.6f\n", sectionsize); 		
		dtprintf("trdump.cc: TrDump::handleevent():Bit length: %.6f\n", bitlength); 		
			
 
		s->dump();	
		int numonline = s->count();
		sectioncount++;

		dtprintf("trdump.cc: TrDump::handleevent(): Number on line: %d\n", numonline); 		
		dtprintf("trdump.cc: TrDump::handleevent(): Checking new section\n"); 		
		rcvr->receive(s, rp);
		if(rp->status == PACKET_ERROR_MUI)
			break;

		s=s->next;

	}		

	totalstatechanges+=(sectioncount -1);

	if(rp->status == PACKET_ERROR_MUI)
	{
		if(steadystate == 1)
		{
			errorpkts++;
			errorbytes+=rp->bytesize;
		}
		dprintf("trgen.cc: TrDump::handleevent(): Packet departing : %d :%d Error\n", rp->id, rp->codeword);
	}
	else
	{
		dprintf("trgen.cc: TrDump::handleevent(): Packet departing : %d : %d Received correctly\n", rp->id, rp->codeword);

	}

	if(steadystate == 1)
	{
		totalpktssent++;	
		totalbytessent+=rp->bytesize;
	}
	write(rp);

	delete pstate;
}

void TrDump::write(Packet *p)
{
	static int count =0;
	//it->mark(p);
	if(count == 100)
	{
		//printf("Cleaning up\n"); 	
		it->clean(p);
		count = 0;
	}
	count++;
	fprintf(fout,  "%.12f %d %d.%d %d.%d %d %d %.12f %.12f\n", sched->currenttime,  p->bytesize,  p->srcaddr, p->srcport, p->dstaddr, p->dstport, p->id, p->status, p->arrivaltime, p->departuretime);
	//printf( "%.12f %d %d.%d %d.%d %d %d\n", sched->currenttime,  p->size,  p->srcaddr, p->srcport, p->dstaddr, p->dstport, p->id, p->status);

}


void TrDump::dumpstats()
{
	double pktthput = (double)(totalpktssent - errorpkts)/(double)((trgen->totalpkts) - (trgen->steadystatepktcnt));
	double avgpkterror = (double) errorpkts/(double)totalpktssent;
	
	stats->totalpkts = totalpktssent;
	stats->noerrorpkts = totalpktssent - errorpkts;

	double bytethput = (double)(totalbytessent - errorbytes)/(double)((trgen->bytecnt));
/*
	printf("Total bytes sent: %d\n", totalbytessent);
	printf("Error bytes: %d\n", errorbytes);
	printf("Bytes count: %d\n", trgen->bytecnt);
*/
	
	stats->totalbytes = totalbytessent;
	stats->noerrorbytes = totalbytessent - errorbytes;

	stats->normpktthput = stats->normoffload * pktthput;
	stats->normbytethput = stats->normoffload * bytethput;
	stats->avgper= avgpkterror;
	
}



