/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <txsched.h>
#include <stateest.h>
#include <statshandler.h>
#include <transmitter.h>

extern Scheduler * sched;
extern InTransit * it;
extern char inputfile[20];
extern Codeset *codeset;
extern TrDump *trdump;
extern TxSched * txsched;
extern StateEstimator * stateest;
extern Transmitter *trn;

extern StatsHandler * stats;
extern int steadystate;

//Maximum distance of a node form the coupler.
double dmax;

//TrGen is the traffic generator class.
//Its function is to generate traffic whoich serves as input to drive the 
//simulation. 
//The traffic generation consists of generating time stamps for packet 
//arrivals to the interfaces of the nodes, packet size generation, 
//source/destination address generation and packet id generation. 
//It is a singleton object.
//It uses a random number generator (rng.cc) to generate different distribtuions. 
//Traffic can also be read from a file.

TrGen::TrGen(int i, int p, double avgpkt, double of, int seed):EventHandler()
{
	debug = debugon("trgen");
	fin = fopen(inputfile, "r");
	if (fin == NULL)
	{
		dtprintf("trgen.cc: TrGen::TrGen(): Could not open input file : %s for reading\n", inputfile);
	}
	iarng = new RNG();
	if(iarng == NULL)
	{
		printf("Error: trgen.cc: Memory allocation failed\n");
		exit(-1);
	}
	psrng = new RNG();
	if(psrng == NULL)
	{
		printf("Error: trgen.cc: Memory allocation failed\n");
		exit(-1);
	}
	distrng = new RNG();
	if(distrng == NULL)
	{
		printf("Error: trgen.cc: Memory allocation failed\n");
		exit(-1);
	}
	
	for(int count =0; count < seed; count++)
	{
		iarng->reset_next_substream();
		psrng->reset_next_substream();
		distrng->reset_next_substream();
	}
	lastarrival = 0;
	done = 0;
	fout = fopen("input.trc", "w");	

	intarrdist = i;
	pktsizedist = p;
	avgpktsize = avgpkt;
	offeredload = of;

//If packet size distribtuion is trimodal with 70% 40 bytes, 20% 1500 bytes and 10% 576 bytes.
//Used to show a pathology when packets sizes have a particular ratio of 
//small size to large size packets.
	if(pktsizedist == TRAFFIC_TM)
	{
		avgpktsize = 0.7*40 + 0.2*1500 + 0.1*576;
	}

	if(pktsizedist == TRAFFIC_BM)
	{
		//For bimodal traffic the value passed as the avgpktsize argument is actually the percentage of short packets.
		fracshortpkt = (double)avgpktsize/100.0; 
		avgpktsize = fracshortpkt * 50 + (1- fracshortpkt) *1000;
	}


//Average inter arrival time is calculated from the offered load and avaerge packet size.

	avginterarrival = (1.0/offeredload) * (double)avgpktsize *8 /(codeset->chiprate * codeset->A);

	pktcnt =0;
	bytecnt=0;
	id = 0;	

}

TrGen::~TrGen()
{
	if(fin != NULL)
		fclose(fin);
}


//The data acquisition function of the TrGen: This function reads the 
//trace file, constructs a packet and processes it.
void TrGen::read()
{
	dtprintf("trgen.cc: TrGen::read(): Entered function\n");


	double arrivaltime;
	int pktid;
	int srcaddr1 = 0;
	int srcaddr2 = 0;
	int srcaddr3 = 0;
	int srcaddr4 = 0;
	int srcport;
	int dstaddr1 = 0;
	int dstaddr2 = 0;
	int dstaddr3 = 0;
	int dstaddr4 = 0;
	int dstport;
	int size;

	double x;

//Read event from trace file
	if(intarrdist == TRAFFIC_FILE)
	{

		//Currently supports Star topology only
		if (fscanf(fin,  "%lf %lf %d %d.%d.%d.%d.%d %d.%d.%d.%d.%d %d\n", &arrivaltime, &x, &size,  &srcaddr1,&srcaddr2, &srcaddr3, &srcaddr4, &srcport,  &dstaddr1, &dstaddr2, &dstaddr3, &dstaddr4, &dstport, &pktid) == EOF)
		{
			//sched->halt = 1;
			dprintf("trgen.cc: TrGen::read(): No more packets.\n");
			dprintf("\n");
			done =1;	
			totalpkts = pktcnt;
			totaltime = lasttime - initialtime;
			offeredload = (totalbits/totaltime);
			stats->normoffload = offeredload/codeset->chiprate;
			printf("Total time: %.12f Total bits: %.6f Offeredload: %.12f\n", totaltime, totalbits, stats->normoffload);

			return;
		}
		if(pktid >= 10000)
		{
			done =1;	
			totalpkts = pktcnt;
			return;
		}
		
		if(pktid== 0)
		{
			initialtime = arrivaltime;
			lasttime = arrivaltime;
		}

		lasttime=arrivaltime;
		totalbits+=size*8.0;

	}
	else
	{
		id++;
		if(pktcnt >= totalpkts)
		{
			dprintf("trgen.cc: TrGen::read(): Done: No more packets %d %d\n", pktcnt, totalpkts);
			done =1;	
			return;
		}
		dprintf("trgen.cc: TrGen::read(): Not done: Number of pkts: %d Total: %d\n", pktcnt, totalpkts);
		arrivaltime = genarrivaltime();
		size = (int) genpacketsize();
		srcaddr4 = dstaddr4 = srcport = dstport = 1;
		pktid = id;

	}


//Call ipport's processing function.

	dprintf("trgen.cc: TrGen::read(): Creating packet\n");
	Packet *packet = new Packet();
	if(packet == NULL)
	{
		printf("Error: trgen.cc: Memory allocation failed\n");
		exit(-1);
	}
	packet->id = pktid;

	packet->srcaddr = (int)(srcaddr4 + srcaddr3* pow(2, 8) + srcaddr2* pow(2, 16) + srcaddr1* pow(2, 24));
	packet->dstaddr = (int)(dstaddr4 + dstaddr3* pow(2, 8) + dstaddr2* pow(2, 16) + dstaddr1 *pow(2, 24));
	packet->srcport = srcport;
	packet->dstport = dstport;
	packet->bytesize = size;

//Initialize the data of the packet.
//Default setting is for binary data (0/1 with p = 0.5)
	//packet->format = DATA_BINARY;
	packet->A = codeset->A;
	packet->N = codeset->N;
	packet->w = codeset->w;

	packet->modulation = modulation;		
	packet->initialize();


	dtprintf("trgen.cc: TrGen::read(): Packetid: %d Src addr: %d Dst addr: %d Size: %d\n", pktid, packet->srcaddr, packet->dstaddr, size);

	dtprintf("trgen.cc: TrGen::read(): Distance: %.6f\n", x);
	dtprintf("trgen.cc: TrGen::read(): Current time: %.12f\n", sched->currenttime);
	dtprintf("trgen.cc: TrGen::read(): Timestamp: %.12f\n", arrivaltime);


	if(size <= 0)
	{
		printf("Error in trace file: Size of packet <= 0: %d \n", pktid);
		exit(-1);
	}

//TBD: Fix format of timestamp (double to long int)
	packet->arrivaltime = arrivaltime;

	if(intarrdist == TRAFFIC_FILE)
	{

		int chosen = 0;
		//When using traffic from file, always assign the same distance to a src addr.
		chosen = packet->srcaddr % (int)numnodes;
		gendistance();
		x = dist[chosen];
		packet->depx = x;

	}
	else
	{
		x = 0;
		int i = gendistance();
		x = dist[i];

		packet->depx = x;
	}
	packet->status = PACKET_WAITINGFORTX;	
	packet->offset = 0;

	int chosen = 0;

//When using traffic from file, always assign the same codeword to a dst addr.
//It does not matter for generated traffic - the destination address and therfore 
//the codeword is chosen uniform randomly from the codeset
	if(intarrdist == TRAFFIC_FILE)
		chosen = packet->dstaddr % codeset->numcodewords;
	else
		chosen = (int) ((random()/(pow(2, 31) - 1))*(codeset->numcodewords));

	packet->codeword = chosen;

	dtprintf("trgen.cc: TrGen::read(): Codeword index: %d\n", chosen);
	dtprintf("trgen.cc: TrGen::read()::Code :");	
	for (int count1 =0; count1 < packet->A; count1++)
	{
		for (int count =packet->N - 1; count >=0 ; count--)
		{
			dprintf("%d ", (codeset->set[(packet->codeword) * (packet->A*packet->N) + packet->N*count1 + count ] ));
		}
	}
	dprintf("\n");


	packet->chiprate = codeset->chiprate;
	packet->bitrate = packet->chiprate/codeset->N;

	double delay  = arrivaltime - sched->currenttime;
	dtprintf("trgen.cc: TrGen::read(): Scheduling packet in %.12f\n", delay);
	dprintf("\n");

	if(intarrdist != TRAFFIC_FILE)
	{
		//Dump the traffic file for future reference/stats calcs
		fprintf(fout,  "%.12f %.6f %d 0.0.0.%d.%d 0.0.0.%d.%d %d\n", arrivaltime, x,  size,  srcaddr4, srcport, dstaddr4, dstport, pktid);
	}


	sched->scheduleevent(this, packet, delay);
	pktcnt++;
	if(steadystate == 1)
		bytecnt+=packet->bytesize;
	if((pktcnt > steadystatepktcnt) && (steadystate != 1))
	{
		sched->scheduleevent(stats, NULL, 0.000001);	
		steadystate =1;
	}
		 
}

//The event handling function of the TrGen. It dequeues a packet and
//sends it to the next component.
void TrGen::handleevent(Packet *p)
{
	dtprintf("trgen.cc: TrGen::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dtprintf("trgen.cc: TrGen::handleevent(): Packet id: %d\n", p->id);
	dtprintf("trgen.cc: TrGen::handleevent(): Packet processing complete, passing it on to the next component.\n");
	dprintf("\n");

	trn->process(p);
	read();
}

void TrGen::dumpstats()
{
	printf("trgen.cc: TrGen::dumpstats():\n");
}

double TrGen::genarrivaltime()
{
	double ia = 0;
	switch(intarrdist)
	{
		case TRAFFIC_EXP:	
			ia = iarng->exponential(avginterarrival);
			break;		
		case TRAFFIC_DET:
			ia = avginterarrival;
			break;
		case TRAFFIC_UNIFORM:
			ia = iarng->uniform(2 * avginterarrival);
			break;
		default:
			printf("Error: Inter-arrival distribution not supported\n");
			exit(-1);
	}

//Round off to the nearest chip boundary 
	ia = ia * codeset->chiprate; 	
	int a = (int)ia;
	ia = (double)a/(codeset->chiprate);
	dtprintf("trgen.cc: TrGen::handleevent(): Departure: %.12f\n", ia);

	double newarrival = lastarrival + ia;	
	lastarrival = newarrival;
	

	
	return(newarrival);		
}

double TrGen::genpacketsize()
{
	double size =0;		
	double p =0.0;
	double shape =0.0;
	double scale =0.0;
	
	switch(pktsizedist)
	{
		case TRAFFIC_EXP:	
			size = psrng->exponential(avgpktsize);
			break;		
		case TRAFFIC_DET:
			size = avgpktsize;
			break;
		case TRAFFIC_UNIFORM:
			size = psrng->uniform(2 *avgpktsize);
			break;
		case TRAFFIC_TM:
			p = (random()/(pow(2, 31) - 1));
			if(p < .7)
				size =40;
			else if((p < .8) && (p > .7))
				size = 576;
			else
				size = 1500;
			break;
		case TRAFFIC_NORMAL:
			size = psrng->normal(avgpktsize, varpktsize);
			break;
		case TRAFFIC_PARETO:
			shape = varpktsize;  
			scale = avgpktsize * ((shape -1) /shape);  //shape must be > 1, usually set to 1.5
			size = psrng->pareto(scale, shape);
			break;
		case TRAFFIC_BM:
			p = (random()/(pow(2, 31) - 1));
			if(p < fracshortpkt)
				size = 50;
			else
				size = 1000;
			break;
		default:
			printf("Error: Packet size distribution not supported\n");
			exit(-1);

	}
	dtprintf("trgen.cc: TrGen::handleevent(): Packet size: %.6f\n", size );
	if (size < 1.0)
		size = 1;
	return(size);

}


int TrGen::gendistance()
{

	static int initial = 0;
	if(initial == 0)
	{
		initial = 1; 
		dist = new double[(int)numnodes];
		if(dist == NULL)
		{
			printf("Error: trgen.cc: Memory allocation failed\n");
			exit(-1);
		}
		for(int count =0; count < numnodes; count++)
		{
			double d =0;
			switch(topologydist)
			{
				case TOPOLOGY_EXP :	
					d = distrng->exponential(avgdistance);
					break;
				case TOPOLOGY_DET:	
					d = avgdistance;
					break;
				case TOPOLOGY_UNIFORM:	
					d = distrng->uniform(2*avgdistance);
					break;
				default:
					printf("Error: Topology distribution not supported\n");
					exit(-1);
			}	
			
//Chip synchronous operation
//Make sure distance is a multiple of 0.02 cm (chip length). This is to avoid handling of cases where
//the distance has a fraction of the chip length. 
//In reality it will be handled by the ranging procedure.

			int d2 = (int)(d * 100.0);
			if(d2 % 2 != 0)
				d2++;
			dist[count] = ((double)d2)/100.0;
			if(dist[count] > dmax)
				dmax = dist[count];
		}
		dtprintf("trgen.cc: TrGen::handleevent(): Dmax: %.6f\n", dmax);


	}			

	int index = (int) ((random()/(pow(2, 31) - 1))*(numnodes));
	return(index);


	//dist = dist * 2.0 * avgdistance/1000.0;
	//printf("Distance: %.15f\n", d);
	
	//double d = distrng->uniform(avgdistance);
	//double dist = 0;	

}
